/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhDFT;
   
DECLARE_NODE(CvfdhDFT)
/*Node
 *
 * @name CvfdhDFT
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhDFT
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 END*/
   
   
class CvfdhDFT : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
      
public:
  CvfdhDFT(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    inOrder = true;
  }      
      
  void cvShiftDFT (CvArr * src_arr, CvArr * dst_arr)
  {
    CvMat *tmp = 0;
    CvMat q1stub, q2stub;
    CvMat q3stub, q4stub;
    CvMat d1stub, d2stub;
    CvMat d3stub, d4stub;
    CvMat *q1, *q2, *q3, *q4;
    CvMat *d1, *d2, *d3, *d4;

    CvSize size = cvGetSize (src_arr);
    CvSize dst_size = cvGetSize (dst_arr);
    int cx, cy;

    if (dst_size.width != size.width || dst_size.height != size.height) {
      cvError (CV_StsUnmatchedSizes, "cvShiftDFT", "Source and Destination arrays must have equal sizes", __FILE__,
	       __LINE__);
    }
    // (9)ץ졼⡼ѤΥƥݥХåե
    if (src_arr == dst_arr) {
      tmp = cvCreateMat (size.height / 2, size.width / 2, cvGetElemType (src_arr));
    }
    cx = size.width / 2;          /* 濴 */
    cy = size.height / 2;

    // (10)14ݸ¤ɽȡΥԡ
    q1 = cvGetSubRect (src_arr, &q1stub, cvRect (0, 0, cx, cy));
    q2 = cvGetSubRect (src_arr, &q2stub, cvRect (cx, 0, cx, cy));
    q3 = cvGetSubRect (src_arr, &q3stub, cvRect (cx, cy, cx, cy));
    q4 = cvGetSubRect (src_arr, &q4stub, cvRect (0, cy, cx, cy));
    d1 = cvGetSubRect (src_arr, &d1stub, cvRect (0, 0, cx, cy));
    d2 = cvGetSubRect (src_arr, &d2stub, cvRect (cx, 0, cx, cy));
    d3 = cvGetSubRect (src_arr, &d3stub, cvRect (cx, cy, cx, cy));
    d4 = cvGetSubRect (src_arr, &d4stub, cvRect (0, cy, cx, cy));

    // (11)ºݤξݸ¤ؤ
    if (src_arr != dst_arr) {
      if (!CV_ARE_TYPES_EQ (q1, d1)) {
	cvError (CV_StsUnmatchedFormats, "cvShiftDFT", "Source and Destination arrays must have the same format",
		 __FILE__, __LINE__);
      }
      cvCopy (q3, d1, 0);
      cvCopy (q4, d2, 0);
      cvCopy (q1, d3, 0);
      cvCopy (q2, d4, 0);
    }
    else {                        /* ץ졼⡼ */
      cvCopy (q3, tmp, 0);
      cvCopy (q1, q3, 0);
      cvCopy (tmp, q1, 0);
      cvCopy (q4, tmp, 0);
      cvCopy (q2, q4, 0);
      cvCopy (tmp, q2, 0);
    }
  }

  void calculate(int output_id, int count, Buffer &out)
  {

    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }

      IplImage* img_in;
      img_in = cvCloneImage(imagePtr->getImage());
    
      IplImage* img_out;

      IplImage *realInput;
      IplImage *imaginaryInput;
      IplImage *complexInput;
      IplImage *image_Re;
      IplImage *image_Im;
      int dft_M, dft_N;
      CvMat *dft_A, tmp;
      double m, M;

      if(img_in->nChannels > 1){
	IplImage *gray;
	gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
	cvCvtColor(img_in, gray, CV_BGR2GRAY);
	img_in = cvCloneImage (gray);
	cvReleaseImage (&gray);
      }

      realInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 1);
      imaginaryInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 1);
      complexInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 2);

      cvScale (img_in, realInput, 1.0, 0.0);
      cvZero (imaginaryInput);
      cvMerge (realInput, imaginaryInput, NULL, NULL, complexInput);

      dft_M = cvGetOptimalDFTSize (img_in->height - 1);
      dft_N = cvGetOptimalDFTSize (img_in->width - 1);
      dft_A = cvCreateMat (dft_M, dft_N, CV_64FC2);
      image_Re = cvCreateImage (cvSize (dft_N, dft_M), IPL_DEPTH_64F, 1);
      image_Im = cvCreateImage (cvSize (dft_N, dft_M), IPL_DEPTH_64F, 1);

      cvGetSubRect (dft_A, &tmp, cvRect (0, 0, img_in->width, img_in->height));
      cvCopy (complexInput, &tmp, NULL);
      if (dft_A->cols > img_in->width) {
	cvGetSubRect (dft_A, &tmp, cvRect (img_in->width, 0, dft_A->cols - img_in->width, img_in->height));
	cvZero (&tmp);
      }

      cvDFT (dft_A, dft_A, CV_DXT_FORWARD, complexInput->height);
      cvSplit (dft_A, image_Re, image_Im, 0, 0);

      cvPow (image_Re, image_Re, 2.0);
      cvPow (image_Im, image_Im, 2.0);
      cvAdd (image_Re, image_Im, image_Re, NULL);
      cvPow (image_Re, image_Re, 0.5);

      cvAddS (image_Re, cvScalarAll (1.0), image_Re, NULL);
      cvLog (image_Re, image_Re);

      cvShiftDFT (image_Re, image_Re);

      cvMinMaxLoc (image_Re, &m, &M, NULL, NULL, NULL);
      cvScale (image_Re, image_Re, 1.0 / (M - m), 1.0 * (-m) / (M - m));

      // Output Setting
      img_out = cvCloneImage (image_Re);

      CvfdImage* image= new CvfdImage(img_out);  
         
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage(&img_in);
      cvReleaseImage(&img_out);
      cvReleaseImage(&realInput);
      cvReleaseImage(&imaginaryInput);
      cvReleaseImage(&complexInput);
      cvReleaseImage(&image_Re);
      cvReleaseImage(&image_Im);

      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* img_in;
	img_in = cvCloneImage(imagePtr.getImage());

	IplImage* img_out;

	IplImage *realInput;
	IplImage *imaginaryInput;
	IplImage *complexInput;
	IplImage *image_Re;
	IplImage *image_Im;
	int dft_M, dft_N;
	CvMat *dft_A, tmp;
	double m, M;

	if(img_in->nChannels > 1){
	  IplImage *gray;
	  gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
	  cvCvtColor(img_in, gray, CV_BGR2GRAY);
	  img_in = cvCloneImage (gray);
	  cvReleaseImage(&gray);
	}

	realInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 1);
	imaginaryInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 1);
	complexInput = cvCreateImage (cvGetSize (img_in), IPL_DEPTH_64F, 2);

	cvScale (img_in, realInput, 1.0, 0.0);
	cvZero (imaginaryInput);
	cvMerge (realInput, imaginaryInput, NULL, NULL, complexInput);

	dft_M = cvGetOptimalDFTSize (img_in->height - 1);
	dft_N = cvGetOptimalDFTSize (img_in->width - 1);
	dft_A = cvCreateMat (dft_M, dft_N, CV_64FC2);
	image_Re = cvCreateImage (cvSize (dft_N, dft_M), IPL_DEPTH_64F, 1);
	image_Im = cvCreateImage (cvSize (dft_N, dft_M), IPL_DEPTH_64F, 1);

	cvGetSubRect (dft_A, &tmp, cvRect (0, 0, img_in->width, img_in->height));
	cvCopy (complexInput, &tmp, NULL);
	if (dft_A->cols > img_in->width) {
	  cvGetSubRect (dft_A, &tmp, cvRect (img_in->width, 0, dft_A->cols - img_in->width, img_in->height));
	  cvZero (&tmp);
	}

	cvDFT (dft_A, dft_A, CV_DXT_FORWARD, complexInput->height);
	cvSplit (dft_A, image_Re, image_Im, 0, 0);

	cvPow (image_Re, image_Re, 2.0);
	cvPow (image_Im, image_Im, 2.0);
	cvAdd (image_Re, image_Im, image_Re, NULL);
	cvPow (image_Re, image_Re, 0.5);

	cvAddS (image_Re, cvScalarAll (1.0), image_Re, NULL);
	cvLog (image_Re, image_Re);

	cvShiftDFT (image_Re, image_Re);

	cvMinMaxLoc (image_Re, &m, &M, NULL, NULL, NULL);
	cvScale (image_Re, image_Re, 1.0 / (M - m), 1.0 * (-m) / (M - m));

	// Output Setting
	img_out = cvCloneImage (image_Re);

	CvfdImage* image= new CvfdImage(img_out);  

	(*outputMapPtr)[it->first] = image;

	cvReleaseImage(&img_in);
	cvReleaseImage(&img_out);
	cvReleaseImage(&realInput);
	cvReleaseImage(&imaginaryInput);
	cvReleaseImage(&complexInput);
	cvReleaseImage(&image_Re);
	cvReleaseImage(&image_Im);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhDFT)
};
