/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

using namespace std;
using namespace FD;

class CvfdhSrcImageToMap;
   
DECLARE_NODE(CvfdhSrcImageToMap)
/*Node
 *
 * @name CvfdhSrcImageToMap
 * @category FDOPENCV_HARK:Misc
 * @description CvfdhSrcImageToMap
 *
 * @input_name IMAGEIN
 * @input_description CvfdImage type input image
 * @input_type CvfdImage
 *
 * @input_name SRCIN
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 * @input_type Vector<ObjectRef>
 *
 * @output_name IMAGEOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Map type output image
 *
 END*/
   
   
class CvfdhSrcImageToMap : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_imageOutID;
      
public:
  CvfdhSrcImageToMap(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcInID = addInput("SRCIN");
    m_imageOutID = addOutput("IMAGEOUT");
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Vector<ObjectRef> > p_sources = getInput(m_srcInID, count);
    const Vector<ObjectRef>& sources = *p_sources;

    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
    if(imagePtr->getImage() == NULL){
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
      return;
    }

    IplImage* img_in;
    img_in = cvCloneImage(imagePtr->getImage());

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);

    for (int i = 0; i < sources.size(); i++) {

      RCPtr<SrcInfo> src = sources[i];

      int width  = src->roi_xy[0];
      int height = src->roi_xy[1];
      int x = src->roi_c[0];
      int y = src->roi_c[1];

      IplImage* tmp_img;
      CvRect rect = cvRect(x - width / 2, y - height / 2, width, height);
      cvSetImageROI(img_in, rect);
      tmp_img = cvCreateImage(cvSize(width, height), img_in->depth, img_in->nChannels);
      cvCopy(img_in, tmp_img);
      cvResetImageROI(img_in);

      CvfdImage* image= new CvfdImage(tmp_img);  
      (*outputMapPtr)[src->id] = ObjectRef(image); 

      cvReleaseImage(&tmp_img);

    }
    
    (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    cvReleaseImage(&img_in);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhSrcImageToMap)
};
